"""
Helper to retry things.

Module inspired by https://github.com/mozilla-releng/redo, but yielding time to sleep instead of sleeping, to use
with asyncio.
"""

import random


def retrier(*, max_attempts, sleeptime, max_sleeptime, sleepscale=1.5, jitter=0.2):
    """Yield time to wait for, after the attempt, if it failed."""
    assert max_attempts > 1
    assert sleeptime >= 0
    assert 0 <= jitter <= sleeptime
    assert sleepscale >= 1

    cur_sleeptime = min(max_sleeptime, sleeptime)

    for attempt in range(max_attempts):
        cur_jitter = random.randint(int(-jitter * 1000), int(jitter * 1000)) / 1000
        yield max(0, cur_sleeptime + cur_jitter)
        cur_sleeptime = min(max_sleeptime, cur_sleeptime * sleepscale)



""" Common HTTP code. """

import asyncio
import logging
import os
import pickle
import urllib.parse

import aiohttp
import appdirs

from sacad import rate_watcher, redo


def aiohttp_socket_timeout(socket_timeout_s):
    """Return a aiohttp.ClientTimeout object with only socket timeouts set."""
    return aiohttp.ClientTimeout(total=None, connect=None, sock_connect=socket_timeout_s, sock_read=socket_timeout_s)


HTTP_NORMAL_TIMEOUT = aiohttp_socket_timeout(12.1)
HTTP_SHORT_TIMEOUT = aiohttp_socket_timeout(6.1)
HTTP_MAX_ATTEMPTS = 3
HTTP_MAX_RETRY_SLEEP_S = 5
HTTP_MAX_RETRY_SLEEP_SHORT_S = 2
DEFAULT_USER_AGENT = "Mozilla/5.0"


class Http:

    """Async HTTP client code."""

    def __init__(
        self,
        *,
        allow_session_cookies=False,
        min_delay_between_accesses=0,
        jitter_range_ms=None,
        rate_limited_domains=None,
        logger=logging.getLogger(),
    ):
        self.allow_session_cookies = allow_session_cookies
        self.session = None
        self.watcher_db_filepath = os.path.join(
            appdirs.user_cache_dir(appname="sacad", appauthor=False), "rate_watcher.sqlite"
        )
        self.min_delay_between_accesses = min_delay_between_accesses
        self.jitter_range_ms = jitter_range_ms
        self.rate_limited_domains = rate_limited_domains
        self.logger = logger

    async def close(self):
        """Close HTTP session to make aiohttp happy."""
        if self.session is not None:
            await self.session.close()

    async def query(  # noqa: C901
        self, url, *, post_data=None, headers=None, verify=True, cache=None, pre_cache_callback=None
    ):
        """
        Send a GET/POST request.

        Send a GET/POST request or get data from cache, retry if it fails, and return a tuple of store in cache
        callback, response content.
        """

        async def store_in_cache_callback():
            pass

        if cache is not None:
            # try from cache first
            if post_data is not None:
                if (url, post_data) in cache:
                    self.logger.debug(f"Got data for URL {url!r} {dict(post_data)} from cache")
                    return store_in_cache_callback, cache[(url, post_data)]
            elif url in cache:
                self.logger.debug(f"Got data for URL {url!r} from cache")
                return store_in_cache_callback, cache[url]

        if self.session is None:
            self._initSession()

        # do we need to rate limit?
        if self.rate_limited_domains is not None:
            domain = urllib.parse.urlsplit(url).path
            rate_limit = domain in self.rate_limited_domains
        else:
            rate_limit = True
        if rate_limit:
            domain_rate_watcher = rate_watcher.AccessRateWatcher(
                self.watcher_db_filepath,
                url,
                self.min_delay_between_accesses,
                jitter_range_ms=self.jitter_range_ms,
                logger=self.logger,
            )

        for attempt, time_to_sleep in enumerate(
            redo.retrier(
                max_attempts=HTTP_MAX_ATTEMPTS, sleeptime=1, max_sleeptime=HTTP_MAX_RETRY_SLEEP_S, sleepscale=1.5
            ),
            1,
        ):
            if rate_limit:
                await domain_rate_watcher.waitAccessAsync()

            try:
                if post_data is not None:
                    async with self.session.post(
                        url,
                        data=post_data,
                        headers=self._buildHeaders(headers),
                        timeout=HTTP_NORMAL_TIMEOUT,
                        ssl=verify,
                    ) as response:
                        content = await response.read()
                else:
                    async with self.session.get(
                        url, headers=self._buildHeaders(headers), timeout=HTTP_NORMAL_TIMEOUT, ssl=verify
                    ) as response:
                        content = await response.read()

                if cache is not None:

                    async def store_in_cache_callback():
                        if pre_cache_callback is not None:
                            # process
                            try:
                                data = await pre_cache_callback(content)
                            except Exception:
                                data = content
                        else:
                            data = content

                        # add to cache
                        if post_data is not None:
                            cache[(url, post_data)] = data
                        else:
                            cache[url] = data

            except (asyncio.TimeoutError, aiohttp.ClientError) as e:
                self.logger.warning(
                    f"Querying {url!r} failed (attempt {attempt}/{HTTP_MAX_ATTEMPTS}): {e.__class__.__qualname_} {e}"
                )
                if attempt == HTTP_MAX_ATTEMPTS:
                    raise
                else:
                    self.logger.debug(f"Retrying in {time_to_sleep:.3f}s")
                    await asyncio.sleep(time_to_sleep)

            else:
                break  # http retry loop

        response.raise_for_status()

        return store_in_cache_callback, content

    async def isReachable(self, url, *, headers=None, verify=True, response_headers=None, cache=None):
        """
        Send a HEAD request.

        Send a HEAD request with short timeout or get data from cache, return True if ressource has 2xx status code,
        False instead.
        """
        if (cache is not None) and (url in cache):
            # try from cache first
            self.logger.debug(f"Got headers for URL {url!r} from cache")
            resp_ok, response_headers = pickle.loads(cache[url])
            return resp_ok

        if self.session is None:
            self._initSession()

        # do we need to rate limit?
        if self.rate_limited_domains is not None:
            domain = urllib.parse.urlsplit(url).path
            rate_limit = domain in self.rate_limited_domains
        else:
            rate_limit = True
        if rate_limit:
            domain_rate_watcher = rate_watcher.AccessRateWatcher(
                self.watcher_db_filepath,
                url,
                self.min_delay_between_accesses,
                jitter_range_ms=self.jitter_range_ms,
                logger=self.logger,
            )

        resp_ok = True
        try:
            for attempt, time_to_sleep in enumerate(
                redo.retrier(
                    max_attempts=HTTP_MAX_ATTEMPTS,
                    sleeptime=0.5,
                    max_sleeptime=HTTP_MAX_RETRY_SLEEP_SHORT_S,
                    sleepscale=1.5,
                ),
                1,
            ):
                if rate_limit:
                    await domain_rate_watcher.waitAccessAsync()

                try:
                    async with self.session.head(
                        url, headers=self._buildHeaders(headers), timeout=HTTP_SHORT_TIMEOUT, ssl=verify
                    ) as response:
                        pass

                except (asyncio.TimeoutError, aiohttp.ClientError) as e:
                    self.logger.warning(
                        f"Probing {url!r} failed (attempt {attempt}/{HTTP_MAX_ATTEMPTS}): {e.__class__.__qualname_} {e}"
                    )
                    if attempt == HTTP_MAX_ATTEMPTS:
                        resp_ok = False
                    else:
                        self.logger.debug(f"Retrying in {time_to_sleep:.3f}s")
                        await asyncio.sleep(time_to_sleep)

                else:
                    response.raise_for_status()

                    if response_headers is not None:
                        response_headers.update(response.headers)

                    break  # http retry loop

        except aiohttp.ClientResponseError as e:
            self.logger.debug(f"Probing {url!r} failed: {e.__class__.__qualname__} {e}")
            resp_ok = False

        if cache is not None:
            # store in cache
            cache[url] = pickle.dumps((resp_ok, response_headers))

        return resp_ok

    async def fastStreamedQuery(self, url, *, headers=None, verify=True):
        """Send a GET request with short timeout, do not retry, and return streamed response."""
        if self.session is None:
            self._initSession()

        response = await self.session.get(
            url, headers=self._buildHeaders(headers), timeout=HTTP_SHORT_TIMEOUT, ssl=verify
        )

        response.raise_for_status()

        return response

    def _buildHeaders(self, headers):
        """Build HTTP headers dictionary."""
        if headers is None:
            headers = {}
        if "User-Agent" not in headers:
            headers["User-Agent"] = DEFAULT_USER_AGENT
        return headers

    def _initSession(self):
        """
        Initialize HTTP session.

        It must be done in a coroutine, see
        https://docs.aiohttp.org/en/stable/faq.html#why-is-creating-a-clientsession-outside-of-an-event-loop-dangerous
        """
        assert self.session is None
        if self.allow_session_cookies:
            cookie_jar = aiohttp.cookiejar.DummyCookieJar()
        else:
            cookie_jar = None
        connector = aiohttp.TCPConnector()
        self.session = aiohttp.ClientSession(connector=connector, cookie_jar=cookie_jar)



""" Additions to the tempfile module. """

import contextlib
import os
import tempfile


@contextlib.contextmanager
def mkstemp(*args, **kwargs):
    """
    Safely generate a temporary file path.

    Context manager similar to tempfile.NamedTemporaryFile except the file is not deleted on close, and only the
    filepath is returned
    """
    fd, filename = tempfile.mkstemp(*args, **kwargs)
    os.close(fd)
    try:
        yield filename
    finally:
        os.remove(filename)



""" Common code for all cover sources cover sources. """

import abc
import asyncio
import itertools
import logging
import operator
import os
import random
import string
import unicodedata
import urllib.parse

import appdirs
import web_cache

from sacad import http_helpers
from sacad.cover import CoverSourceQuality  # noqa: F401

MAX_THUMBNAIL_SIZE = 256


class CoverSource(metaclass=abc.ABCMeta):

    """Base class for all cover sources."""

    def __init__(
        self,
        target_size,
        size_tolerance_prct,
        *,
        min_delay_between_accesses=0,
        jitter_range_ms=None,
        rate_limited_domains=None,
        allow_cookies=False,
    ):
        self.target_size = target_size
        self.size_tolerance_prct = size_tolerance_prct
        self.logger = logging.getLogger(self.__class__.__name__)

        self.http = http_helpers.Http(
            allow_session_cookies=allow_cookies,
            min_delay_between_accesses=min_delay_between_accesses,
            jitter_range_ms=jitter_range_ms,
            rate_limited_domains=rate_limited_domains,
            logger=self.logger,
        )

        self.ua = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:107.0) Gecko/20100101 Firefox/107.0"

        if not hasattr(__class__, "api_cache"):
            db_filepath = os.path.join(appdirs.user_cache_dir(appname="sacad", appauthor=False), "sacad-cache.sqlite")
            os.makedirs(os.path.dirname(db_filepath), exist_ok=True)
            day_s = 60 * 60 * 24
            __class__.api_cache = web_cache.WebCache(
                db_filepath,
                "cover_source_api_data",
                caching_strategy=web_cache.CachingStrategy.FIFO,
                expiration=random.randint(day_s * 7, day_s * 14),  # 1-2 weeks
                compression=web_cache.Compression.DEFLATE,
            )
            __class__.probe_cache = web_cache.WebCache(
                db_filepath,
                "cover_source_probe_data",
                caching_strategy=web_cache.CachingStrategy.FIFO,
                expiration=day_s * 30 * 6,
            )  # 6 months
            logging.getLogger("Cache").debug(
                f"Total size of file {db_filepath!r}: {__class__.api_cache.getDatabaseFileSize()}"
            )
            for cache, cache_name in zip(
                (__class__.api_cache, __class__.probe_cache), ("cover_source_api_data", "cover_source_probe_data")
            ):
                purged_count = cache.purge()
                logging.getLogger("Cache").debug(
                    f"{purged_count} obsolete entries have been removed from cache {cache_name!r}"
                )
                row_count = len(cache)
                logging.getLogger("Cache").debug(f"Cache {cache_name!r} contains {row_count} entries")

    async def closeSession(self):
        """Close HTTP session to make aiohttp happy."""
        await self.http.close()

    async def search(self, album, artist):
        """Search for a given album/artist and return an iterable of CoverSourceResult."""
        self.logger.debug(f"Searching with source {self.__class__.__name__!r}...")
        album = self.processAlbumString(album)
        artist = self.processArtistString(artist)
        url_data = self.getSearchUrl(album, artist)
        if isinstance(url_data, tuple):
            url, post_data = url_data
        else:
            url = url_data
            post_data = None
        try:
            store_in_cache_callback, api_data = await self.fetchResults(url, post_data)
            results = await self.parseResults(api_data)
        except Exception as e:
            # raise
            self.logger.warning(
                f"Search with source {self.__class__.__name__!r} failed: {e.__class__.__qualname__} {e}"
            )
            return ()
        else:
            if results:
                # only store in cache if parsing succeeds and we have results
                await store_in_cache_callback()

        # get metadata
        futures = []
        for result in filter(operator.methodcaller("needMetadataUpdate"), results):
            coroutine = result.updateImageMetadata()
            future = asyncio.ensure_future(coroutine)
            futures.append(future)
        if futures:
            await asyncio.wait(futures)
        for future in futures:
            future.result()  # raise pending exception if any

        # filter
        results_excluded_count = 0
        reference_only_count = 0
        results_kept = []
        for result in results:
            if (
                (result.size[0] + (self.size_tolerance_prct * self.target_size / 100) < self.target_size)
                or (  # skip too small images
                    result.size[1] + (self.size_tolerance_prct * self.target_size / 100) < self.target_size
                )
                or (result.format is None)
                or result.needMetadataUpdate()  # unknown format
            ):  # if still true, it means we failed to grab metadata, so exclude it
                if result.source_quality.isReference():
                    # we keep this result just for the reference, it will be excluded from the results
                    result.is_only_reference = True
                    results_kept.append(result)
                    reference_only_count += 1
                else:
                    results_excluded_count += 1
            else:
                results_kept.append(result)
        result_kept_count = len(results_kept) - reference_only_count

        # log
        self.logger.info(
            f"Got {result_kept_count} relevant ({results_excluded_count + reference_only_count} excluded) results "
            f"from source {self.__class__.__name__!r}"
        )
        for result in itertools.filterfalse(operator.attrgetter("is_only_reference"), results_kept):
            self.logger.debug(
                "%s %s%s %4dx%4d %s%s"
                % (
                    result.__class__.__name__,
                    ("(%02d) " % (result.rank)) if result.rank is not None else "",
                    result.format.name,
                    result.size[0],
                    result.size[1],
                    result.urls[0],
                    " [x%u]" % (len(result.urls)) if len(result.urls) > 1 else "",
                )
            )
        return results_kept

    async def fetchResults(self, url, post_data=None):
        """Get a (store in cache callback, search results) tuple from an URL."""
        if post_data is not None:
            self.logger.debug(f"Querying URL {url!r} {dict(post_data)}...")
        else:
            self.logger.debug(f"Querying URL {url!r}...")
        headers = {}
        self.updateHttpHeaders(headers)
        return await self.http.query(url, post_data=post_data, headers=headers, cache=__class__.api_cache)

    async def probeUrl(self, url, response_headers=None):
        """Probe URL reachability from cache or HEAD request."""
        self.logger.debug(f"Probing URL {url!r}...")
        headers = {}
        self.updateHttpHeaders(headers)
        resp_headers = {}
        resp_ok = await self.http.isReachable(
            url, headers=headers, response_headers=resp_headers, cache=__class__.probe_cache
        )

        if response_headers is not None:
            response_headers.update(resp_headers)

        return resp_ok

    @staticmethod
    def assembleUrl(base_url, params):
        """Build an URL from URL base and parameters."""
        return f"{base_url}?{urllib.parse.urlencode(params)}"

    @staticmethod
    def unaccentuate(s):
        """Replace accentuated chars in string by their non accentuated equivalent."""
        return "".join(c for c in unicodedata.normalize("NFKD", s) if not unicodedata.combining(c))

    @staticmethod
    def unpunctuate(s, *, char_blacklist=string.punctuation):
        """Remove punctuation from string s."""
        # remove punctuation
        s = "".join(c for c in s if c not in char_blacklist)
        # remove consecutive spaces
        return " ".join(filter(None, s.split(" ")))

    #
    # The following methods can or should be overriden in subclasses
    #

    def processQueryString(self, s):
        """Process artist or album string before building query URL."""
        return __class__.unpunctuate(s.lower())

    def processArtistString(self, artist):
        """Process artist string before building query URL."""
        return self.processQueryString(artist)

    def processAlbumString(self, album):
        """Process album string before building query URL."""
        return self.processQueryString(album)

    @abc.abstractmethod
    def getSearchUrl(self, album, artist):
        """
        Build a search results URL from an album and/or artist name.

        If the URL must be accessed with an HTTP GET request, return the URL as a string.
        If the URL must be accessed with an HTTP POST request, return a tuple with:
        - the URL as a string
        - the post parameters as a collections.OrderedDict

        """
        pass

    def updateHttpHeaders(self, headers):
        """Add API specific HTTP headers."""
        pass

    @abc.abstractmethod
    async def parseResults(self, api_data):
        """Parse API data and return an iterable of results."""
        pass



#!/usr/bin/env python3

""" Smart Automatic Cover Art Downloader : search and download music album covers. """

__version__ = "2.7.5"
__author__ = "desbma"
__license__ = "MPL 2.0"

import argparse
import asyncio
import functools
import inspect
import logging
import os
from typing import Any, Optional, Sequence

from sacad import colored_logging, sources
from sacad.cover import (
    HAS_JPEGOPTIM,
    HAS_OPTIPNG,
    HAS_OXIPNG,
    SUPPORTED_IMG_FORMATS,
    CoverImageFormat,
    CoverSourceResult,
)
from sacad.sources.base import CoverSource

COVER_SOURCE_CLASSES = {
    m[0][: -len(CoverSource.__name__)].lower(): m[1]
    for m in inspect.getmembers(sources, lambda x: inspect.isclass(x) and issubclass(x, CoverSource))
}


async def search_and_download(
    album: str,
    artist: str,
    format: CoverImageFormat,
    size: int,
    out_filepath: str,
    *,
    size_tolerance_prct: int,
    amazon_tlds: Sequence[str] = (),
    source_classes: Optional[Sequence[Any]] = None,
    preserve_format: bool = False,
) -> bool:
    """Search and download a cover, return True if success, False instead."""
    logger = logging.getLogger("Main")

    # register sources
    source_args = (size, size_tolerance_prct)
    if source_classes is None:
        source_classes = tuple(COVER_SOURCE_CLASSES.values())
    assert source_classes is not None  # makes MyPy chill
    cover_sources = [cls(*source_args) for cls in source_classes]
    if sources.AmazonCdCoverSource in source_classes:
        for tld in amazon_tlds:
            cover_sources.append(sources.AmazonCdCoverSource(*source_args, tld=tld))

    # schedule search work
    search_futures = []
    for cover_source in cover_sources:
        coroutine = cover_source.search(album, artist)
        future = asyncio.ensure_future(coroutine)
        search_futures.append(future)

    # wait for it
    await asyncio.wait(search_futures)

    # get results
    results = []
    for future in search_futures:
        source_results = future.result()
        results.extend(source_results)

    # sort results
    results = await CoverSourceResult.preProcessForComparison(results, size, size_tolerance_prct)
    results.sort(
        reverse=True,
        key=functools.cmp_to_key(
            functools.partial(
                CoverSourceResult.compare,
                target_size=size,
                size_tolerance_prct=size_tolerance_prct,
            )
        ),
    )
    if not results:
        logger.info("No results")
    else:
        for i, result in enumerate(results, 1):
            logger.debug(f"#{i:02} {result}")

    # download
    done = False
    for result in results:
        try:
            await result.get(
                format,
                size,
                size_tolerance_prct,
                out_filepath,
                preserve_format=preserve_format,
            )
        except Exception as e:
            logger.warning(f"Download of {result} failed: {e.__class__.__qualname__} {e}")
            continue
        else:
            done = True
            break

    # cleanup sessions
    close_cr = []
    for cover_source in cover_sources:
        close_cr.append(cover_source.closeSession())
    await asyncio.gather(*close_cr)

    return done


def setup_common_args(arg_parser: argparse.ArgumentParser) -> None:
    """Set up command line arguments shared between sacad and sacad_r."""
    arg_parser.add_argument(
        "-t",
        "--size-tolerance",
        type=int,
        default=25,
        dest="size_tolerance_prct",
        help="""Tolerate this percentage of size difference with the target size.
                Note that covers with size above or close to the target size will still be preferred
                if available""",
    )
    arg_parser.add_argument(
        "-a",
        "--amazon-sites",
        nargs="+",
        choices=sources.AmazonCdCoverSource.TLDS[1:],
        default=(),
        dest="amazon_tlds",
        help="""Amazon site TLDs to use as search source, in addition to amazon.com""",
    )
    arg_parser.add_argument(
        "-s",
        "--cover-sources",
        choices=tuple(COVER_SOURCE_CLASSES.keys()),
        default=tuple(COVER_SOURCE_CLASSES.keys()),
        nargs="+",
        help="Cover sources to use, if not set use all of them.",
    )
    arg_parser.add_argument(
        "-p",
        "--preserve-format",
        action="store_true",
        default=False,
        help="Preserve source image format if possible. Target format will still be prefered when sorting results.",
    )


def cl_main() -> None:
    """Command line entry point for sacad_r."""
    # parse args
    arg_parser = argparse.ArgumentParser(
        description=f"SACAD v{__version__}. Search and download an album cover.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument("artist", help="Artist to search for")
    arg_parser.add_argument("album", help="Album to search for")
    arg_parser.add_argument("size", type=int, help="Target image size")
    arg_parser.add_argument("out_filepath", help="Output image filepath")
    setup_common_args(arg_parser)
    arg_parser.add_argument(
        "-v",
        "--verbosity",
        choices=("quiet", "warning", "normal", "debug"),
        default="normal",
        dest="verbosity",
        help="Level of logging output",
    )
    args = arg_parser.parse_args()
    args.format = os.path.splitext(args.out_filepath)[1][1:].lower()
    try:
        args.format = SUPPORTED_IMG_FORMATS[args.format]
    except KeyError:
        print(f"Unable to guess image format from extension, or unknown format: {args.format}")
        exit(1)
    args.cover_sources = tuple(COVER_SOURCE_CLASSES[source] for source in args.cover_sources)

    # setup logger
    logging_level = {
        "quiet": logging.CRITICAL + 1,
        "warning": logging.WARNING,
        "normal": logging.INFO,
        "debug": logging.DEBUG,
    }
    logging.getLogger().setLevel(logging_level[args.verbosity])
    if logging_level[args.verbosity] == logging.DEBUG:
        fmt = "%(asctime)s %(levelname)s [%(name)s] %(message)s"
    else:
        fmt = "%(name)s: %(message)s"
    logging_formatter = colored_logging.ColoredFormatter(fmt=fmt)
    logging_handler = logging.StreamHandler()
    logging_handler.setFormatter(logging_formatter)
    logging.getLogger().addHandler(logging_handler)
    if logging_level[args.verbosity] == logging.DEBUG:
        logging.getLogger("asyncio").setLevel(logging.WARNING)
    else:
        logging.getLogger("asyncio").setLevel(logging.CRITICAL + 1)

    # display warning if optipng/oxipng or jpegoptim are missing
    if not HAS_JPEGOPTIM:
        logging.getLogger("Main").warning("jpegoptim could not be found, JPEG crunching will be disabled")
    if not (HAS_OPTIPNG or HAS_OXIPNG):
        logging.getLogger("Main").warning("optipng or oxipng could not be found, PNG crunching will be disabled")

    # search and download
    coroutine = search_and_download(
        args.album,
        args.artist,
        args.format,
        args.size,
        args.out_filepath,
        size_tolerance_prct=args.size_tolerance_prct,
        amazon_tlds=args.amazon_tlds,
        source_classes=args.cover_sources,
        preserve_format=args.preserve_format,
    )
    future = asyncio.ensure_future(coroutine)
    asyncio.get_event_loop().run_until_complete(future)


if __name__ == "__main__":
    cl_main()



""" Discogs cover source. """

import collections
import json

from sacad import __version__
from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource

FUZZY_MODE = False


class DiscogsCoverSourceResult(CoverSourceResult):

    """Discogs search cover result."""

    def __init__(self, *args, **kwargs):
        quality = CoverSourceQuality.NO_UNRELATED_RESULT_RISK
        if FUZZY_MODE:
            quality |= CoverSourceQuality.FUZZY_SEARCH
        else:
            quality |= CoverSourceQuality.EXACT_SEARCH
        super().__init__(*args, source_quality=quality, **kwargs)


class DiscogsCoverSource(CoverSource):

    """
    Cover source using the official API.

    See https://www.discogs.com/developers
    """

    BASE_URL = "https://api.discogs.com"
    API_KEY = "cGWMOYjQNdWYKXDaxVnR"
    API_SECRET = "NCyWcKHWLAvAreyjDdvVogBzVnzPEEDf"  # not that secret in fact

    def __init__(self, *args, **kwargs):
        # https://www.discogs.com/developers#page:home,header:home-rate-limiting
        super().__init__(*args, min_delay_between_accesses=1, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        url_params = collections.OrderedDict()
        if FUZZY_MODE:
            url_params["q"] = f"{artist} - {album}"
        else:
            url_params["artist"] = artist
            url_params["release_title"] = album
        url_params["type"] = "release"
        return __class__.assembleUrl(f"{__class__.BASE_URL}/database/search", url_params)

    def updateHttpHeaders(self, headers):
        """See CoverSource.updateHttpHeaders."""
        headers["User-Agent"] = f"sacad/{__version__}"
        headers["Accept"] = "application/vnd.discogs.v2.discogs+json"
        headers["Authorization"] = f"Discogs key={__class__.API_KEY}, secret={__class__.API_SECRET}"

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        json_data = json.loads(api_data)

        results = []
        for rank, release in enumerate(json_data["results"], 1):
            if release["formats"][0]["name"] != "CD":
                continue
            thumbnail_url = release["thumb"]
            img_url = release["cover_image"]
            url_tokens = img_url.split("/")
            url_tokens.reverse()
            try:
                img_width = int(next(t for t in url_tokens if t.startswith("w:")).split(":", 1)[-1])
                img_height = int(next(t for t in url_tokens if t.startswith("h:")).split(":", 1)[-1])
            except StopIteration:
                continue
            result = DiscogsCoverSourceResult(
                img_url,
                (img_width, img_height),
                CoverImageFormat.JPEG,
                thumbnail_url=thumbnail_url,
                source=self,
                rank=rank,
                check_metadata=CoverImageMetadata.NONE,
            )
            results.append(result)

        return results
